import numpy as np


def row_euclidean_distance(A, B):
    """Euclidean distance between aligned rows of A. An array of length len(A) (==len(B)).

    >>> import numpy as np
    >>> A = np.arange(5 * 16).reshape((5, 16))
    >>> B = 1 + A

    >>> assert all(row_euclidean_distance(A, A) == np.zeros(5))
    >>> assert all(row_euclidean_distance(A, B) == np.array([4., 4., 4., 4., 4.]))

    Note: Not to be confused with the matrix of distances of all pairs of rows. Here, equivalent to the latter diagnonal (see below).

    ```
    from  sklearn.metrics.pairwise import euclidean_distances
    A = np.random.rand(5, 7)
    B = np.random.rand(5, 7)
    assert all(np.diag(euclidean_distances(A, B)) == row_euclidean_distance(A, B))
    ```

    """
    return np.sqrt(((A - B) ** 2).sum(axis=1))


from functools import wraps


# TODO: Add possibility to specify inclusion/exclusion lists
# TODO: If func arg is annotated with "str", don't try to replace it
# TODO: Raise specific error so outside use can catch and handle in the knowledge that replacedments were done.
def resolve_str_specification(obj_store):
    def _resolve_str_specification(func):
        @wraps(func)
        def _func(*args, **kwargs):
            def _args():
                for a in args:
                    if isinstance(a, str) and a in obj_store:
                        yield obj_store[a]
                    else:
                        yield a

            def _kwargs():
                for k, v in kwargs.items():
                    if isinstance(v, str) and v in obj_store:
                        yield k, obj_store[v]
                    else:
                        yield k, v

            return func(*_args(), **dict(_kwargs()))

        return _func

    return _resolve_str_specification



from sklearn.preprocessing import StandardScaler
from gurgle.base import CallableModel




class StdLessStandardScaler(StandardScaler):
    def __init__(self):
        super().__init__(with_std=True)


CallableModel(StdLessStandardScaler)



from typing import TypeVar, Optional, Iterable, Sequence, Callable
from dataclasses import dataclass

from slang.stypes import (
    Waveform,
    Waveforms,
    Chunker,
    Chunk,
    Chunks,
    Featurizer,
    Feature,
    FV,
    FVs,
    Quantizer,
    Snip,
    Snips,
    Snipper)

from slang.chunkers import fixed_step_chunker
from slang import (
    fixed_step_chunker,
    DFLT_CHUNKER,
    DFLT_CHK_SIZE,
    DFLT_WF_TO_SPECTR,
    SpectralProjector,
    DfltWfToChk,
    DfltChkToFv,
    KMeansFvToSnipDist,
)

from lined import Pipeline, iterize


@dataclass
class WfToFv:
    wf_to_chk: Chunker
    chk_to_fv: Featurizer

    def __call__(self, chk):
        return self.chk_to_fv(self.wf_to_chk(chk))


from slang.snip_stats import ClassifiedMomentsFitter
from functools import cached_property


# from slang.util import lazyprop as cached_property


@dataclass
class Gurgle:
    wf_to_chks: Chunker = DfltWfToChk()
    chk_to_fv: Featurizer = SpectralProjector()
    fv_to_snip: Quantizer = KMeansFvToSnipDist()
    snip_stats: ClassifiedMomentsFitter = ClassifiedMomentsFitter

    def __post_init__(self):
        # TODO: Handle all the "if type -- aligned with previous component"
        if isinstance(self.snip_stats, type):
            self.snip_stats = self.snip_stats(n_classes_=self.fv_to_snip.alphabet_size)

    def __call__(self, chk):
        fv = self.chk_to_fv(chk)
        snip, dist = self.fv_to_snip(fv)  # gfv stands for "grouped
        self.snip_stats.fit_partial_single((snip, dist))
        return self.snip_stats((snip, dist))

    @cached_property
    def wf_to_fvs(self):
        return Pipeline(self.wf_to_chks, iterize(self.chk_to_fv))



from lined.tools import generator_version
from lined import Pipeline
from functools import partial

import numpy as np


class DFLT:
    gain = 10
    cutoff = 50
    dispstr = '*'
    prefix = '|'
    fv_num_sep = '\n'


def viz_norm(x, gain=10, cutoff=50):
    return min(cutoff, x * gain)


def ascii_levels(x, viz_norm=viz_norm, dispstr=DFLT.dispstr, prefix=DFLT.prefix):
    return prefix + dispstr * int(viz_norm(x))


dflt_num_to_viz_str = ascii_levels


def multipart_fv_str(fv,
                     num_to_viz_str=dflt_num_to_viz_str,
                     fv_num_sep=DFLT.fv_num_sep):
    return fv_num_sep.join(map(dflt_num_to_viz_str, dflt_chk_to_fv(fv))) + '\n'


ascii_levels_fv_to_str = partial(multipart_fv_str,
                                 num_to_viz_str=ascii_levels,
                                 fv_num_sep=DFLT.fv_num_sep)

raw_fv_to_str = partial(multipart_fv_str,
                        num_to_viz_str=str,
                        fv_num_sep=DFLT.fv_num_sep)

dflt_fv_to_str = ascii_levels_fv_to_str

int16_max = np.iinfo(np.int16).max


def log_max_chk_to_fv(chk: np.ndarray, gain=50):
    """A chk_to_fv that has to do with volume"""
    return [np.log(max(chk.max(), abs(chk.min()), 1)) / np.log(int16_max) * gain]
    # return np.log(min(1, max(np.abs(chk)))) / np.log(int16_max) * gain


dflt_chk_to_fv = log_max_chk_to_fv


def chk_to_fv_viz(chk: np.ndarray,
                  chk_to_fv=dflt_chk_to_fv,
                  fv_to_str=dflt_fv_to_str):
    """Takes a chk and returns a string representing the fv. Meant to be displayed dynamically"""
    fv = chk_to_fv(chk)
    fv_str = fv_to_str(fv)
    return fv_str


from IPython.core.display import clear_output


def refresh_and_print(iterator_output):
    clear_output(wait=True)
    n, viz = iterator_output
    print(n, viz, sep='\n')


dflt_output_callback = refresh_and_print


def make_and_consume_iterator(iterator_func, *interator_func_args, **interator_func_kwargs):
    """make an consume and iterator"""
    for _ in iterator_func(*interator_func_args, **interator_func_kwargs):
        pass


# import time
#
# for x in range(0, 5):
#     b = "Loading" + "." * x
#     print(b, end="\r")
#     time.sleep(1)

from taped import LiveWf, WfChunks

from itertools import islice


def chk_std(chk):
    return np.std(chk)


same_line_print = partial(print, end='\r')

from typing import Optional


def launch(chk_to_fv=chk_std,
           fv_to_viz=print,
           input_device_index: Optional[int] = None,
           sr: int = 44100,
           chk_size: int = 2048,
           **kwargs
           ):
    """

    :param chk_to_fv:
    :param fv_to_viz:
    :return:
    """
    print("Starting...")
    with WfChunks(input_device_index=input_device_index,
                  sr=sr,
                  chk_size=chk_size,
                  **kwargs) as wf_chunks:
        try:
            for chk in wf_chunks:
                fv = chk_to_fv(chk)
                if fv is not None:
                    fv_to_viz(fv)

        except KeyboardInterrupt:
            print('KeyboardInterrupt: Okay... closing down...')

    print("... closing shop.")


same_line_print = partial(print, end='\r')


def log_print(fv):
    print(np.log2(1 + fv))


def exp_print(fv, e=2):
    print(fv ** e)


#### Gurgle ###############################
from functools import partial
from slang.spectrop import mk_chk_fft
from slang.chunkers import simple_fixed_step_chunker
from lined import Pipeline, iterize

chk_size = 2048
wf_to_chks = partial(simple_fixed_step_chunker, chk_size=chk_size)
chk_to_spectr = mk_chk_fft(chk_size)

from slang.spectrop import logarithmic_bands_matrix

M = logarithmic_bands_matrix(7)
bands_sp = partial(np.dot, b=M.T)

wf_to_spectr = Pipeline(wf_to_chks,
                        iterize(chk_to_spectr),
                        iterize(bands_sp))

from gurgle.sp_gurgle import PartialFitPipeline, ResidueGurgle, IncrementalPCA, ScaledIncrementalPCA
from lined import Pipeline
import operator

gurgle = ResidueGurgle(projector=ScaledIncrementalPCA(n_components=2))


def compute_outlier(chk):
    # fv = wf_to_spectr(chk)
    spectr = chk_to_spectr(chk)
    # return gurgle(spectr)
    fv, r = gurgle(spectr)
    if r is not None:
        return max(0, 2 + r[0]) ** 2
    else:
        return None


def print_ascii_levels_2_70(fv):
    print(ascii_levels(fv, viz_norm=partial(viz_norm, gain=2, cutoff=70)))


def print_ascii_levels_01_70(fv):
    print(ascii_levels(fv, viz_norm=partial(viz_norm, gain=0.1, cutoff=70)))


_allowed_objs = [
    chk_std,
    print,
    compute_outlier,
    print_ascii_levels_2_70,
    print_ascii_levels_01_70,
]


def allowed_objects():
    g = globals()
    for obj in _allowed_objs:
        if isinstance(obj, str) and obj in g:
            yield obj, g[obj]
        else:
            yield obj.__name__, obj


# def allowed_objects():
#     for k, v in globals().items():
#         if callable(v):
#             yield k, v

obj_store = dict(allowed_objects())


def main():
    import argh
    from gurgle.util import resolve_str_specification

    def list_possible_inputs():
        print(*obj_store, sep='\n')

    print("""
    Example usages:
    
    To get a list of elements:
        gurgle-terminal list-possible-inputs
    For volume tracking:
        gurgle-terminal launch chk_std print_ascii_levels_01_70
    For online spectral projector outliers tracking:
        gurgle-terminal launch compute_outlier print_ascii_levels_2_70

    """)

    argh.dispatch_commands([
        resolve_str_specification(obj_store)(launch),
        list_possible_inputs
    ])


if __name__ == '__main__':
    main()



import os
import importlib
import site
import inspect


def loaded_module_from_dotpath_and_filepath(dotpath, filepath):
    """Get module object from file path and module dotpath"""
    module_spec = importlib.util.spec_from_file_location(dotpath, filepath)
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module


def module_path(module):
    try:
        _module_path = module.__path__[0]
    except:
        _module_path = inspect.getfile(module)
        if _module_path.endswith('__init__.py'):
            _module_path = module_path[:len('__init__.py')]
    return _module_path


def submodules(module, on_error='print', prefix=None):
    from py2store.filesys import FileCollection

    prefix = prefix or site.getsitepackages()[0]

    _module_path = module_path(module)
    assert os.path.isdir(_module_path), f"Should be a directory: {module_path}"

    for filepath in FileCollection(_module_path, '{f}.py', max_levels=0):
        dotpath = '.'.join(filepath[len(prefix):(-len('.py'))].split(os.path.sep))
        if dotpath.startswith('.'):
            dotpath = dotpath[1:]
        if not dotpath.endswith('__init__'):
            try:
                yield loaded_module_from_dotpath_and_filepath(dotpath, filepath)
            except Exception as e:
                if on_error == 'print':
                    print(f"{e} ({dotpath}: {filepath})")
                elif on_error == 'raise':
                    raise


def objects_of_module(module, max_levels=0):
    for a in dir(module):
        if not a.startswith('_'):
            yield getattr(module, a)
    if os.path.isdir(module_path(module)):
        if max_levels > 0:
            for submodule in submodules(module):
                yield from objects_of_module(submodule, max_levels - 1)


def obj_to_dotpath(obj):
    return f"{obj.__module__}.{obj.__name__}"


def finding_objects_of_module_with_given_methods(module, method_names=None, max_levels=1):
    module_dotpath = module.__name__

    objects = {obj_to_dotpath(obj): obj for obj in
               filter(lambda o: isinstance(o, type) and o.__module__.startswith(module_dotpath),
                      objects_of_module(module, max_levels))}

    if method_names is None:
        return objects
    else:
        if isinstance(method_names, str):
            method_names = [method_names]
        method_names = set(method_names)

        return {dotpath: obj for dotpath, obj in objects.items() if method_names.issubset(dir(obj))}



from functools import wraps, cached_property
import numpy as np
from scipy.spatial.distance import cdist

from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import IncrementalPCA, PCA
from sklearn.utils.validation import check_is_fitted, check_array
from sklearn.base import TransformerMixin, BaseEstimator

from gurgle.base import CallableModel, TransparentModel


def instantiate_if_type(obj):
    if isinstance(obj, type):
        return obj()
    else:
        return obj


def orth(A):
    """
    Orthonormalize the matrix A.

    Note: Code is more complicated than the purest mathematical need because of numerical problem mitigation.

    :param A: a n-by-k array
    :return: a n-by-k array whose row spans the same space as the rows of A and with the normal rows

    >>> M = np.array([[1,2,3], [3,4,5]])
    >>> O = orth(M)
    >>> # checking the product of O and its transpose is the identity
    >>> np.allclose(np.dot(O, O.T), np.eye(2))
    True

    """
    u, s, vh = np.linalg.svd(A.T, full_matrices=False)
    N, M = A.shape
    eps = np.finfo(float).eps
    tol = max(np.max(M), np.max(N)) * np.amax(s) * eps
    num = np.sum(s > tol, dtype=int)
    Q = u[:, :num]
    return Q.T


def space_residue(self, X):
    return X - np.dot(X, self.orth_projection)


def sub_space_residue(self, X):
    return np.dot(self.space_residue(X), self.orth_scaling)


class ScaledPCA(PCA):
    scaler_cls = StandardScaler
    scaler_ = None

    @wraps(PCA.fit)
    def fit(self, X, y=None, *args, **kwarg):
        self.scaler_ = self.scaler_cls().fit(X)
        super().fit(self.scaler_.transform(X), y=None, *args, **kwarg)
        return self

    @wraps(PCA.transform)
    def transform(self, X):
        return super().transform(self.scaler_.transform(X))

    @wraps(PCA.inverse_transform)
    def inverse_transform(self, X):
        return self.scaler_.inverse_transform(super().inverse_transform(X))


class ProjectorResiduesOnCall(CallableModel):
    def __init__(self, model=ScaledPCA, call_method='transform', score_scaler=TransparentModel):
        super().__init__(model, call_method)
        self.score_scaler = instantiate_if_type(score_scaler)
        self._x_buffer = list()

    # @cached_property
    # def projection(self):
    #     return np.dot(self.scalings_, self.scalings_.T)
    #
    # @cached_property
    # def orth_projection(self):
    #     return orth(self)

    def is_fit(self):
        return hasattr(self.model, 'components_') and self.model.components_ is not None

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector."""
        # TODO: Make this independent of self.model.components_ (should only use X and self.model.transform(X))
        #   so that ScaledPCA can work as well.
        if hasattr(self.model, 'scaler_'):  # TODO: Get rid of this hack by learning some damn linear algebra!
            return np.sum(cdist(self.model.scaler_.transform(X), self.model.components_), axis=1).ravel()
        else:
            return np.sum(cdist(X, self.model.components_), axis=1).ravel()

    def fit(self, X, y=None, *args, **kwargs):
        super().fit(X, y, *args, **kwargs)
        scores = self.decision_function(X)
        self.score_scaler.fit(scores.reshape((-1, 1)))
        return self

    def postproc(self, X, model_output):
        fv, scaled_score = None, None
        if self.is_fit():
            scores = self.decision_function(X)
            scaled_scores = self.score_scaler.transform(scores.reshape((-1, 1)))
            fv, scaled_score = X[0], scaled_scores[0]
        return fv, scaled_score


class IncrementalProjectorResiduesOnCall(ProjectorResiduesOnCall):
    def __init__(self, model=IncrementalPCA, call_method='transform', score_scaler=StandardScaler, learn_batch_size=1):
        super().__init__(model, call_method)
        self.score_scaler = score_scaler
        self.learn_batch_size = learn_batch_size
        self._x_buffer = list()

    def preproc(self, x):
        self._x_buffer.append(x)
        return np.array([x])

    def postproc(self, X, model_output):
        fv, scaled_score = super().postproc(X, model_output)
        if len(self._x_buffer) >= self.learn_batch_size:
            self.model.partial_fit(np.array(self._x_buffer))  # learn a batch
            self._x_buffer = list()  # reset buffer
        return fv, scaled_score


def partial_fittable_instance(obj, name=None, **kwargs):
    """Asserts that obj has a partial_fit, and instantiates it with **kwargs if obj is given as a type"""
    if name is None:
        if isinstance(obj, type):
            name = obj.__name__
        else:
            name = type(obj).__name__

    assert hasattr(obj, 'partial_fit'), f"{name} doesn't have a partial_fit: {obj} "
    if isinstance(obj, type):
        obj = obj(**kwargs)
    return obj


class ScaledIncrementalPCA(IncrementalPCA, ScaledPCA):
    scaler_cls = StandardScaler
    scaler_ = None

    def is_fit(self):
        return hasattr(self, 'components_') and self.components_ is not None

    @wraps(IncrementalPCA.partial_fit)
    def partial_fit(self, X, y=None, check_input=True):
        if self.scaler_ is None:
            self.scaler_ = self.scaler_cls()

        if not self.is_fit():
            # then this is the first training data piece we see, so we need to make sure we have enough to start fitting
            nrows, fv_size = X.shape
            if self.n_components is None:
                self.n_components = fv_size
            X = complete_with_fake_data_for_warmup(minimum_n_rows_to_fit=self.n_components, X=X)
        self.scaler_.partial_fit(X)
        scaled_X = self.scaler_.transform(X)
        super().partial_fit(scaled_X, y, check_input=check_input)
        return self


class ResidueGurgle(BaseEstimator):
    learn_batch_size = None

    def __init__(self, projector=ScaledIncrementalPCA, score_scaler=StandardScaler):
        # assert n_components is not None, "You need to specify an actual number of components. No None allowed here."
        self.projector = partial_fittable_instance(projector, 'projector')
        self.score_scaler = partial_fittable_instance(score_scaler, 'score_scaler')

        self._x_buffer = list()
        self.learn_batch_size = self.learn_batch_size or (self.n_components + 1)

    @property
    def n_components(self):
        return self.projector.n_components

    def is_fit(self):
        return hasattr(self.projector, 'components_') and self.projector.components_ is not None

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
        """
        # check_is_fitted(self, ['components_'])
        # X = check_array(X)
        # return np.sum(cdist(self.projector, self.projector.components_), axis=1).ravel()
        # TODO: Make this independent of self.model.components_ (should only use X and self.model.transform(X))
        #   so that ScaledPCA can work as well.
        if hasattr(self.projector, 'scaler_'):  # TODO: Get rid of this hack by learning some damn linear algebra!
            return np.sum(cdist(self.projector.scaler_.transform(X), self.projector.components_), axis=1).ravel()
        else:
            return np.sum(cdist(X, self.projector.components_), axis=1).ravel()

    def partial_fit(self, X, y=None):
        if not self.is_fit():
            # then this is the first training data piece we see, so we need to make sure we have enough to start fitting
            nrows, fv_size = X.shape
            if self.n_components is None:
                self.projector.n_components = fv_size
            X = complete_with_fake_data_for_warmup(minimum_n_rows_to_fit=self.n_components, X=X)
        self.projector.partial_fit(X, y)
        scores = self.decision_function(X)
        self.score_scaler.partial_fit(scores.reshape((-1, 1)))
        return self

    def __call__(self, x):
        fv, scaled_score = None, None
        X = np.array([x])
        self._x_buffer.extend(X.tolist())
        if self.is_fit():
            fvs = self.projector.transform(X)
            scores = self.decision_function(X)
            scaled_scores = self.score_scaler.transform(scores.reshape((-1, 1)))
            fv, scaled_score = fvs[0], scaled_scores[0]
        if len(self._x_buffer) >= (self.learn_batch_size or len(x)):
            self.partial_fit(np.array(self._x_buffer))  # learn a batch
            self._x_buffer = list()  # reset buffer
        return fv, scaled_score

    #
    # def __call__(self, x):
    #     x = self.preproc(x)
    #     model_output = self.call_method(x)
    #     post_output = self.postproc(x, model_output)
    #     return post_output

def complete_with_fake_data_for_warmup(minimum_n_rows_to_fit, X=None, fv_size=None):
    """Makes fake data to warmup a partial fit process.
    If no X is given, will return a random minimum_n_rows_to_fit x fv_size matrix (with values between 0 and 1)
    If X is given, will repeat the rows in a cycle until the minimum_n_rows_to_fit is reached

    >>> X = complete_with_fake_data_for_warmup(3, fv_size=2);
    >>> X.shape
    (3, 2)
    >>> import numpy as np
    >>> complete_with_fake_data_for_warmup(5, X=np.array([[1,2,3], [4,5,6]]))
    array([[1, 2, 3],
           [4, 5, 6],
           [1, 2, 3],
           [4, 5, 6],
           [1, 2, 3]])
    """
    if X is None:
        assert fv_size is not None, "You need to have some data, or specify an fv_size"
        return np.random.rand(minimum_n_rows_to_fit, fv_size)
    else:
        nrows, fv_size = X.shape
    missing_n_rows = max(0, minimum_n_rows_to_fit - nrows)
    if missing_n_rows > 0:
        return np.array(X.tolist() * int(1 + np.ceil(missing_n_rows / nrows)))[:minimum_n_rows_to_fit]
    else:
        return X


############## Not used at the moment ##################################################################################


class GurgleWrap(BaseEstimator):
    learn_batch_size = None

    def __init__(self, model, score_scaler=StandardScaler, learn_batch_size=1):
        # assert n_components is not None, "You need to specify an actual number of components. No None allowed here."
        self.model = partial_fittable_instance(model, 'model')
        self.score_scaler = partial_fittable_instance(score_scaler, 'score_scaler')

        self._x_buffer = list()
        self.learn_batch_size = learn_batch_size

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
        """
        # check_is_fitted(self, ['components_'])
        X = check_array(X)
        return np.sum(cdist(X, self.projector.components_), axis=1).ravel()

    def partial_fit(self, X, y=None):
        if not self.is_fit():
            # then this is the first training data piece we see, so we need to make sure we have enough to start fitting
            nrows, fv_size = X.shape
            if self.n_components is None:
                self.projector.n_components = fv_size
            X = complete_with_fake_data_for_warmup(minimum_n_rows_to_fit=self.n_components, X=X)
        self.projector.partial_fit(X, y)
        scores = self.decision_function(X)
        self.score_scaler.partial_fit(scores.reshape((-1, 1)))
        return self

    def __call__(self, x):
        fv, scaled_score = None, None
        X = np.array([x])
        self._x_buffer.extend(X.tolist())
        if self.is_fit():
            fvs = self.projector.transform(X)
            scores = self.decision_function(X)
            scaled_scores = self.score_scaler.transform(scores.reshape((-1, 1)))
            fv, scaled_score = fvs[0], scaled_scores[0]
        if len(self._x_buffer) >= (self.learn_batch_size or len(x)):
            self.partial_fit(np.array(self._x_buffer))  # learn a batch
            self._x_buffer = list()  # reset buffer
        return fv, scaled_score


class PartialFitPipeline:
    def __init__(self, *steps):
        self.steps = steps
        self.last_step = steps[-1]

    def __getattr__(self, attr):
        """Delegate method to wrapped store if not part of wrapper store methods"""
        return getattr(self.last_step, attr)

    def transform_until_last_step(self, X):
        for step in self.steps[:-1]:
            X = step.transform(X)
        return X

    def partial_fit_transform(self, X, y=None, **kwargs):
        for step in self.steps:
            X = step.partial_fit(X, y, **kwargs).transform(X)
        return X

    def partial_fit(self, X, y=None, **kwargs):
        self.partial_fit_transform(X, y, **kwargs)
        return self

    def fit(self, X, y=None, **kwargs):
        for step in self.steps[:-1]:
            X = step.fit(X, y, **kwargs).transform(X)
        self.last_step.fit(X)
        return self

    def transform(self, X):
        return self.last_step.transform(self.transform_until_last_step(X))

    def predict_proba(self, X):
        return self.last_step.predict_proba(self.transform_until_last_step(X))

    def predict(self, X):
        return self.last_step.predict(self.transform_until_last_step(X))


class ProjectionDecisionFuncMixin:
    components_ = None  # lint appeaser

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.

        The anomaly score of an input sample is computed based on different
        detector algorithms. For consistency, outliers are assigned with
        larger anomaly scores.

        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.

        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        check_is_fitted(self, ['components_'])
        X = check_array(X)
        return np.sum(cdist(X, self.components_), axis=1).ravel()

# class ScaledPCA(PCA):
#     scaler_cls = StandardScaler
#     scaler_ = None
#
#     @wraps(PCA.fit)
#     def fit(self, X, y=None, *args, **kwarg):
#         self.scaler_ = self.scaler_cls().fit(X)
#         super().fit(self.scaler_.transform(X), y=None, *args, **kwarg)
#         return self
#
#     @wraps(PCA.transform)
#     def transform(self, X):
#         return super().transform(self.scaler_.transform(X))
#
#     @wraps(PCA.inverse_transform)
#     def inverse_transform(self, X):
#         return self.scaler_.inverse_transform(super().inverse_transform(X))


# class PcaGurgle(ScaledIncrementalPCA, ProjectionDecisionFuncMixin):
#     learn_batch_size = None
#
#     @wraps(ScaledIncrementalPCA.__init__)
#     def __init__(self, n_components, score_scaler=StandardScaler, *args, **kwargs):
#         # assert n_components is not None, "You need to specify an actual number of components. No None allowed here."
#         super().__init__(n_components, *args, **kwargs)
#         self.score_scaler = partial_fittable_instance(score_scaler, 'score_scaler')
#         self._x_buffer = list()
#         self.learn_batch_size = self.learn_batch_size or self.n_components
#
#     def is_fit(self):
#         return hasattr(self, 'components_') and self.components_ is not None
#
#     def partial_fit(self, X, y=None, check_input=True):
#         if not self.is_fit():
#             # then this is the first training data piece we see, so we need to make sure we have enough to start fitting
#             nrows, fv_size = X.shape
#             if self.n_components is None:
#                 self.n_components = fv_size
#             X = complete_with_fake_data_for_warmup(minimum_n_rows_to_fit=self.n_components, X=X)
#         super().partial_fit(X, y, check_input)
#         scores = self.decision_function(X)
#         self.score_scaler.partial_fit(scores.reshape((-1, 1)))
#         return self
#
#     def __call__(self, x):
#         fv, scaled_score = None, None
#         X = np.array([x])
#         self._x_buffer.extend(X.tolist())
#         if self.is_fit():
#             fvs = self.transform(X)
#             scores = self.decision_function(X)
#             scaled_scores = self.score_scaler.transform(scores.reshape((-1, 1)))
#             fv, scaled_score = fvs[0], scaled_scores[0]
#         if len(self._x_buffer) >= (self.learn_batch_size or len(x)):
#             self.partial_fit(np.array(self._x_buffer))  # learn a batch
#             self._x_buffer = list()  # reset buffer
#         return fv, scaled_score

# if self._x_buffer
# X = np.array([x])  # .resize((1, -1))
# if self.is_fit():
#     fvs = self.transform(X)
# else:
#     fvs = self.partial_fit(X).transform(X)
# scores = self.decision_function(X)
# scaled_scores = self.score_scaler.transform(scores.reshape((-1, 1)))
# self.partial_fit(X)
# return fvs[0], scaled_scores[0]



"""
To run, run this in the terminal:

```
streamlit run streamlit_app.py
```
"""
import streamlit as st
# To make things easier later, we're also importing numpy and pandas for
# working with sample data.
import numpy as np
import pandas as pd

st.title('My first app')

# st.write("Here's our first attempt at using data to create a table:")
# st.write(pd.DataFrame({
#     'first column': [1, 2, 3, 4],
#     'second column': [10, 20, 30, 40]
# }))
#
# df = pd.DataFrame({
#     'another col': ['apples', 'and', 'bananas'],
#     'and yet another': [[1, 2], 'yum', float]
# })
#
# df  # lone litteral will result in a st.write call
#
# df.T;

import matplotlib.pyplot as plt
import numpy as np
import streamlit as st
import time

# max_x = 5
# max_y = 10

from itertools import cycle
from collections import deque

# def cyclic_data_gen(max_items=100):
#     c = cycle(range(max_y))
#     d = deque(range(max_y), max_x)
#     for i in range(max_items):
#         d.append(next(c))
#         yield d

from itertools import tee, count, chain, islice


def cyclic_data_gen(max_items=100, max_y=10):
    yield from islice(cycle(range(max_y)), 0, max_items)


def prepend_several(iterator, fillval=None, n_prepends=1):
    """Prepend a single value in front of an iterator
    >>> list(prepend_several([2, 3, 4], 0, n_prepends=3))
    [0, 0, 0, 2, 3, 4]
    """
    "Prepend a single value in front of an iterator"
    return chain([fillval] * n_prepends, iterator)


def sliding_window(iterable, n=2):
    """Iterable of sliding windows
    >>> list(sliding_window(range(7), 3))
    [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]
    """
    iterables = tee(iterable, n)

    for iterable, num_skipped in zip(iterables, count()):
        for _ in range(num_skipped):
            next(iterable, None)

    return zip(*iterables)


def streaming_graph(data_gen, window_length=5, inter_tick_sleep=0.01, max_y=10):
    sliding_window_intervals = sliding_window(prepend_several(data_gen, 0, window_length - 1), window_length)

    # sliding_window_intervals = sliding_window(data_gen, window_length)
    fig, ax = plt.subplots()

    x = np.arange(0, window_length)
    ax.set_ylim(0, max_y)
    line, = ax.plot(x, next(sliding_window_intervals), '-o')
    the_plot = st.pyplot(plt)

    def init():  # give a clean slate to start
        line.set_ydata([np.nan] * len(x))

    init()

    i = 0
    for interval in sliding_window_intervals:
        line.set_ydata(interval)
        the_plot.pyplot(plt)
        time.sleep(inter_tick_sleep)
        i += 1


from taped import LiveWf, WfChunks

with WfChunks() as wf_chunks:
    def gen():
        for chk in islice(wf_chunks, 0, 200):
            fv = np.std(chk)
            print(fv)
            yield fv


    streaming_graph(gen(), window_length=20, max_y=800)

    # streaming_graph(cyclic_data_gen(20, max_y=10))

#
# chart_data = pd.DataFrame(
#     np.random.randn(20, 3),
#     columns=['a', 'b', 'c'])
#
# st.line_chart(chart_data)
#
# import streamlit as st
# import pandas as pd
# import time
# import altair as alt
# from altair import Chart, X, Y, Axis, SortField, OpacityValue
#
#
# # ---------------------------------------------------------------#
# # Creating an empty chart in the beginning when the page loads
# # ---------------------------------------------------------------#
# bars = alt.Chart(data).mark_bar().encode(
#     x=X('1:Q', axis=Axis(title='ATP Ranking Points'),
#         y=Y('0:Q', axis=Axis(title='The Big Four'))
#         ).properties(
#         width=650,
#         height=400
#     ))
# # This global variable 'bar_plot' will be used later on
# bar_plot = st.altair_chart(bars)
#
#
# x = st.slider('Select the year range',1996, 2017, (1996, 2017))

# def plot_bar_animated_altair(df, week):
#     bars = alt.Chart(df, title="Ranking as of week :" + week).encode(
#         x=X('ranking_points:Q', axis=Axis(title='ATP Ranking Points'),
#             y=Y('full_name:N', axis=Axis(title='The Big Four'), sort='-x'),
#             color=alt.Color('full_name:N'),
#             .properties(
#             width=650,
#             height=400
#         )))
#
#     if st.button('Cue Chart'):
#         for week in week_list:
#     # weekly_df -> this dataframe (sample shown above) contains
#     # data for a particular week which is passed to
#     # the 'plot_bar_animated_altair' function.
#     # week -> Current week title, eg:-  2016-06-10
#
#             bars = plot_bar_animated_altair(weekly_df, week)
#             time.sleep(0.01)
#
#             bar_plot.altair_chart(bars)
#
#             st.balloons()  # Displays some celebratory balloons for glamour!


print('To run: streamlit run streamlit_app.py')



"""
To run, run this in the terminal:

```
streamlit run streamlit_app.py
```
"""
import streamlit as st
# To make things easier later, we're also importing numpy and pandas for
# working with sample data.
import numpy as np
import pandas as pd

st.title('My first app')

# st.write("Here's our first attempt at using data to create a table:")
# st.write(pd.DataFrame({
#     'first column': [1, 2, 3, 4],
#     'second column': [10, 20, 30, 40]
# }))
#
# df = pd.DataFrame({
#     'another col': ['apples', 'and', 'bananas'],
#     'and yet another': [[1, 2], 'yum', float]
# })
#
# df  # lone litteral will result in a st.write call
#
# df.T;

import matplotlib.pyplot as plt
import numpy as np
import streamlit as st
import time

# max_x = 5
# max_y = 10

from itertools import cycle
from collections import deque

# def cyclic_data_gen(max_items=100):
#     c = cycle(range(max_y))
#     d = deque(range(max_y), max_x)
#     for i in range(max_items):
#         d.append(next(c))
#         yield d

from itertools import tee, count, chain, islice


def cyclic_data_gen(max_items=100, max_y=10):
    yield from islice(cycle(range(max_y)), 0, max_items)


def prepend_several(iterator, fillval=None, n_prepends=1):
    """Prepend a single value in front of an iterator
    >>> list(prepend_several([2, 3, 4], 0, n_prepends=3))
    [0, 0, 0, 2, 3, 4]
    """
    "Prepend a single value in front of an iterator"
    return chain([fillval] * n_prepends, iterator)


def sliding_window(iterable, n=2):
    """Iterable of sliding windows
    >>> list(sliding_window(range(7), 3))
    [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]
    """
    iterables = tee(iterable, n)

    for iterable, num_skipped in zip(iterables, count()):
        for _ in range(num_skipped):
            next(iterable, None)

    return zip(*iterables)


def streaming_graph(data_gen, window_length=5, inter_tick_sleep=0.01, max_y=10):
    sliding_window_intervals = sliding_window(prepend_several(data_gen, 0, window_length - 1), window_length)

    # sliding_window_intervals = sliding_window(data_gen, window_length)
    fig, ax = plt.subplots()

    x = np.arange(0, window_length)
    ax.set_ylim(0, max_y)
    line, = ax.plot(x, next(sliding_window_intervals), '-o')
    the_plot = st.pyplot(plt)

    def init():  # give a clean slate to start
        line.set_ydata([np.nan] * len(x))

    init()

    i = 0
    for interval in sliding_window_intervals:
        line.set_ydata(interval)
        the_plot.pyplot(plt)
        time.sleep(inter_tick_sleep)
        i += 1


from taped import LiveWf, WfChunks


with WfChunks() as wf_chunks:
    def gen(x):
        for chk in islice(wf_chunks, 0, x):
            fv = np.std(chk)
            yield fv
    try:
        for fv in gen(100):
            print(fv)

    except KeyboardInterrupt:
        print('KeyboardInterrupt: Okay... closing down...')












from functools import wraps

import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator

# TODO: Make identity_func "identifiable". If we use the following one, we can use == to detect it's use,
# TODO: ... but there may be a way to annotate, register, or type any identity function so it can be detected.
def identity_func(x):
    return x

static_identity_method = staticmethod(identity_func)


class NoSuchItem:
    pass


no_such_item = NoSuchItem()


def cls_wrap(cls, obj):
    if isinstance(obj, type):

        @wraps(obj, updated=())
        class Wrap(cls):
            @wraps(obj.__init__)
            def __init__(self, *args, **kwargs):
                wrapped = obj(*args, **kwargs)
                super().__init__(wrapped)

        # Wrap.__signature__ = signature(obj)

        return Wrap
    else:
        return cls(obj)


class CallableModel:
    _call_method_str = 'transform'

    def __init__(self, model):
        if isinstance(model, type):
            model = model()
        self.model = model
        self.call_method = getattr(self.model, self._call_method_str)

    def __getattr__(self, attr):
        """Delegate method to wrapped store if not part of wrapper store methods"""
        return getattr(self.model, attr)

    def __dir__(self):
        return list(set(dir(self.__class__)).union(self.model.__dir__()))  # to forward dir to delegated stream

    wrap = classmethod(cls_wrap)

    def fit(self, X, y=None, *args, **kwargs):
        self.model.fit(X, y, *args, **kwargs)
        return self

    @staticmethod
    def preproc(self, x):
        return np.array([x])

    @staticmethod
    def postproc(self, x, model_output):
        return model_output[0]

    def __call__(self, x):
        x = self.preproc(x)
        model_output = self.call_method(x)
        post_output = self.postproc(x, model_output)
        return post_output


class TransparentModel(TransformerMixin, BaseEstimator):
    def __init__(self, **kwargs):
        pass

    def fit(self, X, y=None, *args, **kwargs):
        return self

    def partial_fit(self, X, y=None, *args, **kwargs):
        return self

    def transform(self, X):
        return X

    def predict_proba(self, X):
        return X

    def predict(self, X):
        return np.nan * np.zeros(X.shape[0])
